// array.cpp

#include "jcpplisp.h"

extern CELL *NILCell,*TCell,*CHARACTER,*COLDIRECTION,*COLINPUT,*COLOUTPUT,*STRING,*LIST,*FUNCTION,*MACRO;

extern char lineStr[BUFLEN];

CELL *cfunction::jcons(CELL *car,CELL *cdr)
{
    CELL *c;

    c = getCell();
    c->attr = 'l';
    c->car = car;
    c->cdr = cdr;
    return c;
}

CELL *cfunction::jnth(long n,CELL *c)
{
    if (c->attr != 'l') return NULL;
    if (n == 0L) return c->car;
    return jnth(n - 1,c->cdr);
}

CELL *cfunction::jprenth(long n,CELL *c)
{
    if (c->attr != 'l') return NULL;
    if (n == 0L) return c;
    return jprenth(n - 1,c->cdr);
}

long cprint::jnlength(CELL *c)
{
    long i = 0;

    for (;;) {
	if (c->attr != 'l') return i;
	c = c->cdr;
	i++;
    }
}

CELL *cfunction::SetArray(CELL *top,SYMVAL *loc,long locn,long lock)
{
    CELL *c;

    if ((top->cdr)->attr != 'l') goto serror;
    IncCellStack(NULL); // debug on 5 Jan. 2017
    IncCellStack(top);  // debug on 5 Jan. 2017
    c = eval((top->cdr)->car,loc,locn,lock);
    PopCellStack();     // debug on 5 Jan. 2017
    if (c == NULL) goto serror;
    if (c->attr != 'l') goto serror;
    c->attr = 'a';
    return c;

  serror:
    printf("エラー: 配列型に設定できません。\n");
    errorlisp(loc,locn);
    return NULL;
}


CELL *cfunction::ArrayCar(CELL *a)
{
    if (a->attr != 'a') { puts("配列ではありません"); return NULL; }
    return a->car;
}

CELL *cfunction::ArrayCdr(CELL *a)
{
    if (a->attr != 'a') { puts("配列ではありません"); return NULL; }
    return a->cdr;
}

CELL *cmemory::get1char(char *m)
{
    unsigned char c1,c2;

    c1 = (unsigned char)*m;
    if (isKanji(&c1)) c2 = (unsigned char)*(m + 1);
    else {
	c2 = (unsigned char)NULL;
	if (c1 == '\\') {
	    if (*(m + 1) == '\\') ;
	    else if (*(m + 1) == '"') {
		c1 = '"';
	    }
	    else {
		c1 = *(m + 1);
	    }
	}
    }
    return getMojiCell(c1,c2);
}

unsigned char *ctextproc::getRestchar(char *m)
{
    unsigned char c1,c2;

    c1 = (unsigned char)*m;
    if (isKanji(&c1)) m++;
    else {
	if (c1 == '\\') {
	    if (*(m + 1) == '\\') m++;
	    else if (*(m + 1) == '"') {
		m++;
	    }
	    else {
		m++;
	    }
	}
    }
    m++;
    return (unsigned char *)m;
}

CELL *cmemory::getMojiretsuCell(char *m)
{
    CELL *p,*i,*c,*head,*q;
    AINT *k;

    i = getCell();
    i->attr = 'i';
    k = getInt(1L);
    i->car = (CELL *)k;

    head = p = getCell();
    p->attr = 'a';
    p->car = i;
    q = getCell();
    q->car = CHARACTER;
    p->cdr = q;
    m++;
    for (;;) {
	p = q;
	if (*m == '"') break;
	c = get1char(m);
	m = (char *)getRestchar(m);
	q = getCell();
	q->attr = 'l';
	q->car = c;
	p->cdr = q;
    }
    p = NILCell;
    if (p == NULL) return NULL; // dummy
    return head;
}

BOOL cfunction::toPathMojiretsu(CELL *a,unsigned char *name,int maxlen)
{
    int i;

    if (a->attr != 'a') return FALSE;
    if ((a->car)->attr != 'i') return FALSE;
    if (((AINT *)((a->car)->car))->f != 1L) return FALSE;
    a = a->cdr;
    if (a->car != CHARACTER) return FALSE;
    a = a->cdr;
    if (jnlength(a) < 1) return FALSE;
    i = 0;
    for (;;) {
	if (i >= maxlen) return FALSE;
	if (jnlength(a) == 0) break;
	if (a->attr != 'l') return FALSE;
	if ((a->car)->attr != 'm') return FALSE;
	if ((a->car)->cdr == NULL) name[i] = (unsigned char)(long)((a->car)->car);
	else {
	    name[i++] = (unsigned char)(long)((a->car)->car);
	    if (i >= maxlen) return FALSE;
	    name[i] = (unsigned char)(long)((a->car)->cdr);
	}
	i++;
	a = a->cdr;
    }
    name[i] = '\0';
    return TRUE;
}

void cprint::PutString(CELL *a,FILE *fp)  // a = (#\h #\e #\l #\l #\o) etc.
{
    fprintf(fp,"\"");
    for (;;) {
	if (jnlength(a) == 0) break;
	if (a->attr != 'l') return;
	if ((a->car)->attr != 'm') return;
	if ((a->car)->cdr == NULL) {
	    if ((unsigned char)(long)((a->car)->car) == '"') fprintf(fp,"\\");
	    else if ((unsigned char)(long)((a->car)->car) == '\\') fprintf(fp,"\\");
	    fprintf(fp,"%c",(unsigned char)(long)((a->car)->car));
	}
	else {
	    fprintf(fp,"%c",(unsigned char)(long)((a->car)->car));
	    fprintf(fp,"%c",(unsigned char)(long)((a->car)->cdr));
	}
	a = a->cdr;
    }
    fprintf(fp,"\"");
}

void cprint::PutStringC(CELL *a,FILE *fp)  // a = (#\h #\e #\l #\l #\o) etc.
{
    for (;;) {
	if (jnlength(a) == 0) break;
	if (a->attr != 'l') return;
	if ((a->car)->attr != 'm') return;
	if ((a->car)->cdr == NULL) {
	    fprintf(fp,"%c",(unsigned char)(long)((a->car)->car));
	}
	else {
	    fprintf(fp,"%c",(unsigned char)(long)((a->car)->car));
	    fprintf(fp,"%c",(unsigned char)(long)((a->car)->cdr));
	}
	a = a->cdr;
    }
}

CELL *cfunction::topen(CELL *p,SYMVAL *loc,long locn,long lock)
{
    char pathname[PATH_LEN + 1];
    char mode[5];
    FILE *fp;
    CELL *q;

    if (p->attr != 'l') goto serror;

    IncCellStack(NULL);
    IncCellStack(p);

    if ((q = eval(p->car,loc,locn,lock)) == NULL) { PopCellStack(); goto serror; }

    PopCellStack();

    if (q->attr != 's' && q->attr != 'a') goto serror;
    if (q->attr == 's') { strncpy(pathname,(char *)(q->car),PATH_LEN); pathname[PATH_LEN] = '\0'; }
    else if (q->attr == 'a') {
	if(!toPathMojiretsu(q,(unsigned char *)pathname,PATH_LEN)) goto serror;
    }
    p = p->cdr;
    if (p->attr != 'l') {
	strcpy(mode,"rb");
    }
    else if (p->car == COLDIRECTION) {
	p = p->cdr;
        if (p->attr != 'l') strcpy(mode,"rb");
        else {
            q = eval(p->car,loc,locn,lock);
	    if (q == NULL) goto serror;
            if (q == COLINPUT) strcpy(mode,"rb");
	    else if (q == COLOUTPUT) strcpy(mode,"wb");
            else strcpy(mode,"rb");
	}
    }
    else strcpy(mode,"rb");
    fp = fopen(pathname,mode);
    if (fp == NULL) {
	printf("エラー: ファイルを開くことができません。\n");
	errorlisp(loc,locn);
	/*return NILCell;*/
	return NULL;
    }
    //return SetFilePointer(fp);
    q = SetFilePointer(fp);
    if (q == NILCell) {
	puts("エラー: ファイルの数が多すぎます");
	errorlisp(loc,locn);
	return NULL;
    }
    return q;
serror:
    //return NILCell;
    errorlisp(loc,locn);
    return NULL;
}

CELL *cmemory::ReadLine(void)
{
    char *p;
    CELL *q,*r,*r2;

    fgets(lineStr + 1,BUFLEN - 1,stdin);
    p = lineStr;
    *p++ = '"';
    while (*p != '\0') {
	if (*p == '\n') { *p = '\0'; break; }
	else if (*p == '\\') { p++; slide(p,1); *p++ = '\\'; }
	else if (*p == '"') { *p++ = '\\'; slide(p,1); *p++ = '"'; }
	else if (isKanji((unsigned char *)p)) { p += 2; }
	else p++;
    }
    strcat(lineStr,"\"");
    q = getMojiretsuCell(lineStr);

    r = getCell();
    r2 = getCell();
    r->attr = 'l';
    r->car = q;
    r->cdr = r2;
    r2->attr = 'l';
    r2->car = NILCell;
    r2->cdr = NILCell;
    q->mtop = r;
    return q;
}

CELL *cfunction::StringEqual(CELL *a1,CELL *a2)
{
    if (a1->attr != 'a' || a2->attr != 'a') return NILCell;
    if ((a1->car)->attr != 'i' || (a2->car)->attr != 'i') return NILCell;
    if (((AINT *)((a1->car)->car))->f != 1 || ((AINT *)((a2->car)->car))->f != 1) return NILCell;
    a1 = a1->cdr;
    a2 = a2->cdr;
    if (a1->car != CHARACTER || a2->car != CHARACTER) return NILCell;
    a1 = a1->cdr;
    a2 = a2->cdr;
    for (;;) {
	if (a1->attr != 'l') {
	    if (a2->attr != 'l') return TCell;
	    else return NILCell;
	}
	if (a2->attr != 'l') {
	    if (a1->attr != 'l') return TCell;
	    else return NILCell;
	}
	if ((unsigned char)(long)((a1->car)->car) != (unsigned char)(long)((a2->car)->car)) return NILCell;
	if ((unsigned char)(long)((a1->car)->cdr) != (unsigned char)(long)((a2->car)->cdr)) return NILCell;
	a1 = a1->cdr;
	a2 = a2->cdr;
    }
}

BOOL cfunction::isAllMoji(CELL *p)
{
    for (;;) {
	if (p->attr != 'l') {
	    if (p != NILCell) return FALSE;
	    else return TRUE;
	}
	if ((p->car)->attr != 'm') return FALSE;
	p = p->cdr;
    }
}

CELL *cfunction::coerce(CELL *p,SYMVAL *loc,long locn,long lock)
{
    CELL *x,*y,*q,*i,*head;
    AINT *k;

    if (p->attr != 'l') goto serror;
    x = p->car;

    IncCellStack(NULL);
    IncCellStack(p);

    x = eval(x,loc,locn,lock);

    PopCellStack();

    if (x == NULL) return NULL;
    p = p->cdr;
    if (p->attr != 'l') goto serror;
    y = p->car;

    IncCellStack(NULL);
    IncCellStack(x);

    y = eval(y,loc,locn,lock);

    PopCellStack();

    if (y == NULL) return NULL;
    if (y == STRING) {
	if (x->attr != 'l' && x != NILCell) goto serror;
	if (!isAllMoji(x)) goto serror;
	i = getCell();
	i->attr = 'i';
	k = getInt(1L);
	i->car = (CELL *)k;
	head = p = getCell();
	p->attr = 'a';
	p->car = i;
	q = getCell();
	q->car = CHARACTER;
	p->cdr = q;
	q->cdr = Copy(x);
	return head;
    }
    else if (y == LIST) {
	if (x->attr != 'a' || (x->car)->attr != 'i' || ((AINT *)((x->car)->car))->f != 1L) goto serror;
	x = x->cdr;
	if (x->car != CHARACTER) goto serror;
	return Copy(x->cdr);
    }
    else if (y == FUNCTION) return x;

serror:
    printf("エラー: COERCE: 文法が間違っています。\n");
    errorlisp(loc,locn);
    return NULL;
}

CELL *cfunction::tchar(CELL *p,SYMVAL *loc,long locn,long lock)
{
    CELL *q,*r;
    long k,i,len;
    long cnt;
    char *ch;

    if (p->attr != 'l') goto serror;
    q = p->car;

    IncCellStack(NULL);
    IncCellStack(p);

    q = eval(q,loc,locn,lock);

    PopCellStack();

    if (q == NULL) return NULL;
    p = p->cdr;
    if (p->attr != 'l') goto serror;
    r = p->car;

    IncCellStack(NULL);
    IncCellStack(q);

    r = eval(r,loc,locn,lock);

    PopCellStack();

    if (r == NULL) return NULL;
    if (q->attr != 'a') goto serror;
    if (r->attr != 'i') goto serror;
    k = ((AINT *)(r->car))->f;
    if (k < 0) goto serror;
    if ((q->car)->attr != 'i' || ((AINT *)((q->car)->car))->f != 1L) goto serror;
    q = q->cdr;
    if (q->car != CHARACTER) goto serror;
    q = q->cdr;
    cnt = 0;
    for (;;) {
	if (q->attr != 'l') return NILCell;
	if (cnt == k) return q->car;
	q = q->cdr;
	cnt++;
    }

serror:
    printf("エラー: 文法が間違っています。\n");
    errorlisp(loc,locn);
    return NULL;
}

unsigned long cmathproc::XorShift()
{
    static unsigned long x = 123456789,y = 362436069,z = 521288629,w = 88675123;
    unsigned long t;

    t = x ^ (x << 11);
    x = y; y = z; z = w;
    return w = (w ^ (w >> 19)) ^ (t ^ (t >> 8));
}

CELL *cfunction::Random(CELL *max,SYMVAL *loc,long locn,long lock)
{
    CELL *p,*q;
    long m,k;
    AINT *i;
    long double m2,k2;
    AFLOAT *i2;

    p = eval(max,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'i') {
	m = ((AINT *)(p->car))->f;
	if (m < 0) {
	    puts("エラー: 負数は指定できません");
	    errorlisp(loc,locn);
	    return NULL;
	}
	k = XorShift() / 10;
	k = (long)(k * ((long double)m / XORSHIFT_MAX));
	q = getCell();
	i = getInt(k);
	q->attr = 'i';
	q->car = (CELL *)i;
	return q;
    }
    else if (p->attr == 'n') {
	m2 = ((AFLOAT *)(p->car))->f;
	if (m2 < 0.0) {
	    puts("エラー: 負数は指定できません");
	    errorlisp(loc,locn);
	    return NULL;
	}
	k = XorShift() / 10;
	k2 = k * (m2 / XORSHIFT_MAX);
	q = getCell();
	i2 = getNumber(k2);
	q->attr = 'n';
	q->car = (CELL *)i2;
	return q;
    }
    else {
	puts("エラー: 数値ではありません");
	return NULL;
    }
}

CELL *cfunction::catch_lisp(CELL *p,SYMVAL *loc,long locn)
{
    CELL *tag_name;
    CELL *s;

    if (p->attr != 'l') {
	puts("エラー: CATCHの引数が異常です\n");
	errorlisp(loc,locn);
	return NULL;
    }
    s = p->cdr;
    IncCellStack(NULL);
    IncCellStack(s);
    tag_name = eval(p->car,loc,locn,0);
    PopCellStack();
    if (tag_name == NULL) {
	puts("エラー: CATCHのタグの名前の評価に失敗しました\n");
	return NULL;
    }
    p = NILCell;
    IncCellStack(NULL);
    IncCellStack(tag_name);
    IncCellStack(s);
    for (;;) {
	if (s->attr != 'l') { PopCellStack(); return p; }
	if ((p = eval(s->car,loc,locn,0)) == NULL) {
	    PopCellStack();
	    if (throw_flag && tag_name == throw_name) {
		throw_flag = FALSE;
		return throw_value;
	    }
	    return NULL;
	}
	s = s->cdr;
    }
}

CELL *cfunction::throw_lisp(CELL *p,SYMVAL *loc,long locn)
{
    CELL *s;

    if (p->attr != 'l') {
	puts("エラー: THROWの引数が異常です\n");
	errorlisp(loc,locn);
	return NULL;
    }
    s = p->cdr;
    IncCellStack(NULL);
    IncCellStack(s);
    throw_name = eval(p->car,loc,locn,0);
    PopCellStack();
    if (throw_name == NULL) {
	puts("エラー: THROWのタグの名前の評価に失敗しました\n");
	return NULL;
    }
    if (s->attr != 'l') {
	puts("エラー: THROWの引数が異常です\n");
	errorlisp(loc,locn);
	return NULL;
    }
    IncCellStack(NULL);
    IncCellStack(throw_name);
    throw_value = eval(s->car,loc,locn,0);
    PopCellStack();
    throw_flag = TRUE;
    return NULL;
}

CELL *cfunction::unwind_protect(CELL *p,SYMVAL *loc,long locn)
{
    CELL *q,*ret,*ret2;

    if (p->attr != 'l') {
	puts("エラー: UNWIND-PROTECTの引数が異常です\n");
	errorlisp(loc,locn);
	return NULL;
    }
    q = p->cdr;
    if (q->attr != 'l' && q != NILCell) {
	puts("エラー: UNWIND-PROTECTの引数が異常です\n");
	errorlisp(loc,locn);
	return NULL;
    }
    IncCellStack(NULL);
    IncCellStack(q);
    ret = eval(p->car,loc,locn,0);
    PopCellStack();
    IncCellStack(NULL);
    IncCellStack(ret);
    ret2 = tprogn(q,loc,locn,0);
    PopCellStack();
    if (ret2 == NULL) {
	throw_flag = FALSE;
	return NULL;
    }
    return ret;
}

CELL *cfunction::symbol_function(CELL *p,SYMVAL *loc,long locn)
{
    SYMVAL *q;
    CELL *a;

    if (p->attr != 'l') {
	puts("エラー: SYMBOL-FUNCTIONの引数が異常です\n");
	errorlisp(loc,locn);
	return NULL;
    }
    p = eval(p->car,loc,locn,0);
    if (p == NULL) return NULL;
    q = getFuncval2(p);
    if (q == NULL) {
	puts("エラー: システム関数、または未定義の関数またはマクロです\n");
	errorlisp(loc,locn);
	return NULL;
    }
    a = getCell();
    a->attr = 'l';
    if (q->attr == 'f') a->car = FUNCTION;
    else a->car = MACRO;
    a->cdr = q->val;
    return a;
}

CELL *cfunction::block_lisp(CELL *p,SYMVAL *loc,long locn)
{
    CELL *tag_name;
    CELL *s;

    if (p->attr != 'l') {
	puts("エラー: BLOCKの引数が異常です\n");
	errorlisp(loc,locn);
	return NULL;
    }
    s = p->cdr;
//    IncCellStack(NULL);
//    IncCellStack(s);
    tag_name = p->car;
//    PopCellStack();
    if (tag_name->attr != 's') {
	puts("エラー: BLOCKの名前が異常です");
	errorlisp(loc,locn);
	return NULL;
    }
    p = NILCell;
    IncCellStack(NULL);
    IncCellStack(tag_name);
    IncCellStack(s);
    for (;;) {
	if (s->attr != 'l') { PopCellStack(); return p; }
	if ((p = eval(s->car,loc,locn,0)) == NULL) {
	    PopCellStack();
	    if (block_throw_flag && tag_name == block_throw_name && block_throw_value != NULL) {
		block_throw_flag = FALSE;
		return block_throw_value;
	    }
	    return NULL;
	}
	s = s->cdr;
    }
}

CELL *cfunction::return_from_lisp(CELL *p,SYMVAL *loc,long locn)
{
    CELL *s;

    if (p->attr != 'l') {
	puts("エラー: RETURN-FROMの引数が異常です\n");
	errorlisp(loc,locn);
	return NULL;
    }
    s = p->cdr;
//    IncCellStack(NULL);
//    IncCellStack(s);
    block_throw_name = p->car;
//    PopCellStack();
    if (block_throw_name->attr != 's') {
	puts("エラー: RETURN-FROMの名前が異常です");
	errorlisp(loc,locn);
	return NULL;
    }
    if (s->attr != 'l') {
//	puts("エラー: RETURN-FROMの引数が異常です\n");
//	errorlisp(loc,locn);
	block_throw_value = NILCell;
	block_throw_flag = TRUE;
	return NULL;
    }
    IncCellStack(NULL);
    IncCellStack(throw_name);
    block_throw_value = eval(s->car,loc,locn,0);
    PopCellStack();
    block_throw_flag = TRUE;
    return NULL;
}

// End of file

