// math.cpp

#include "jcpplisp.h"
/*#include "dummy.h"*/

extern CELL *NILCell,*TCell;

long double CellToNumber(CELL *p)
{
    if (p->attr != 'n') return 0.0;
    return ((AFLOAT *)(p->car))->f;
}

CELL *cmemory::NumberToCell(long double v)
{
    CELL *p;

    p = getCell();
    p->attr = 'n';
    p->car = (CELL *)getNumber(v);
    return p;
}

long cmemory::CellToInt(CELL *p)
{
    if (p->attr != 'i') return 0;
    return ((AINT *)(p->car))->f;
}

CELL *cmemory::IntToCell(long v)
{
    CELL *p;

    p = getCell();
    p->attr = 'i';
    p->car = (CELL *)getInt(v);
    return p;
}

long cmemory::CellToBunshi(CELL *p)
{
    if (p->attr != 'b') return 0;
    return ((AINT *)(p->car))->f;
}

long cmemory::CellToBunbo(CELL *p)
{
    if (p->attr != 'b') return 0;
    return ((AINT *)(p->cdr))->f;
}

CELL *cmemory::BunsuToCell(long i,long o)
{
    CELL *p;

    p = getCell();
    p->attr = 'b';
    p->car = (CELL *)getInt(i);
    p->cdr = (CELL *)getInt(o);
    return p;
}

CELL *cfunction::jSqrt(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long double v;
    long n;
    long vd,vu;
    long i,k;
    BOOL bunshiInt;
    long bunshi;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'n') {
	v = CellToNumber(p);
	if (v < 0.0) {
	    puts("エラー: 負数は平方根はとれません");
	    errorlisp(loc,locn);
	    return NULL;
	}
	v = sqrt((double)v);
	return NumberToCell(v);
    }
    else if (p->attr == 'i') {
	n = CellToInt(p);
	if (n < 0) {
	    puts("エラー: 負数は平方根はとれません");
	    errorlisp(loc,locn);
	    return NULL;
	}
	v = sqrt((double)n);
	vd = (long)floor((double)v);
	vu = (long)ceil((double)v);
	k = vd * vd;
	if (k == n) return IntToCell(vd);
	k = vu * vu;
	if (k == n) return IntToCell(vu);
	return NumberToCell(v);
    }
    else if (p->attr == 'b') {
	n = CellToBunshi(p);
	if (n < 0) {
	    puts("エラー: 負数は平方根はとれません");
	    errorlisp(loc,locn);
	    return NULL;
	}
	v = (long double)sqrt((double)n);
	vd = (long)floor((double)v);
	vu = (long)ceil((double)v);
	k = vd * vd;
	if (k == n) {
	    bunshi = vd;
	    bunshiInt = TRUE;
	    if (bunshi == 0) return IntToCell(0);
	}
	else {
	    k = vu * vu;
	    if (k == n) {
		bunshi = vu;
		bunshiInt = TRUE;
		if (bunshi == 0) return IntToCell(0);
	    }
	    else {
		bunshiInt = FALSE;
	    }
	}

	if (bunshiInt) {
	    n = CellToBunbo(p);
	    if (n <= 0) {
		puts("エラー: 負数は平方根はとれません");
		errorlisp(loc,locn);
		return NULL;
	    }
	    v = sqrt((double)n);
	    vd = (long)floor((double)v);
	    vu = (long)ceil((double)v);
	    k = vd * vd;
	    if (k == n) return BunsuToCell(bunshi,vd);
	    k = vu * vu;
	    if (k == n) return BunsuToCell(bunshi,vu);
	    return NumberToCell(bunshi / v);
	}
	else {
	    v = sqrt((long double)(CellToBunshi(p) / (long double)CellToBunbo(p)));
	    return NumberToCell(v);
	}
    }
    else {
	puts("エラー: 数値以外の平方根はとれません");
	errorlisp(loc,locn);
	return NULL;
    }
}

CELL *cmemory::njExpt(long double x,CELL *y)
{
    switch (y->attr) {
	case 'n':
	    return NumberToCell(pow(x,CellToNumber(y)));
	case 'i':
	    return NumberToCell(pow(x,(long double)CellToInt(y)));
	case 'b':
	    return NumberToCell(pow(x,(long double)CellToBunshi(y) / (long double)CellToBunbo(y)));
	default:
	    puts("エラー: 指数部は数値でなければなりません");
	    return NULL;
    }
}

CELL *cmemory::ijExpt(long x,CELL *y)
{
    long i,n,k;
    long double d;
    long dd,du;
    long limit;

    switch (y->attr) {
	case 'n':
	    return NumberToCell(pow(x,CellToNumber(y)));
	case 'i':
	    k = 1;
	    n = CellToInt(y);
	    if (x != 0L) limit = LONG_MAX / x;
	    else limit = LONG_MAX;
	    for (i = 0;i < n;i++) {
		if (k >= limit) {
		    puts("エラー: 整数値がオーバーフローしました");
		    return NULL;
		}
		k *= x;
	    }
	    return IntToCell(k);
	case 'b':
	    d = pow(x,1.0 / (long double)CellToBunbo(y));
	    dd = (long)floor(d);
	    du = (long)ceil(d);
	    k = 1L;
	    n = CellToBunbo(y);
	    for (i = 0;i < n;i++) k *= dd;
	    if (x == k) {
		k = 1L;
		n = CellToBunshi(y);
		limit = LONG_MAX / dd;
		for (i = 0;i < n;i++) {
		    if (k >= limit) {
			puts("エラー: 整数値がオーバーフローしました");
			return NULL;
		    }
		    k *= dd;
		}
		return IntToCell(k);
	    }
	    k = 1L;
	    for (i = 0;i < n;i++) k *= du;
	    if (x == k) {
		k = 1L;
		n = CellToBunshi(y);
		limit = LONG_MAX / du;
		for (i = 0;i < n;i++) {
		    if (k >= limit) {
			puts("エラー: 整数値がオーバーフローしました");
			return NULL;
		    }
		    k *= du;
		}
		return IntToCell(k);
	    }
	    return NumberToCell(pow(x,(long double)CellToBunshi(y) / (long double)CellToBunbo(y)));
	default:
	    puts("エラー: 指数部は数値でなければなりません");
	    return NULL;
    }
}

CELL *cmemory::bjExpt(long a,long b,CELL *y)
{
    long shi,bo,i,n;
    long double d;
    long dd,du;
    long k,k2;
    long limit;

    switch (y->attr) {
	case 'i':
	    n = CellToInt(y);
	    shi = 1L;
	    if (a != 0L) limit = LONG_MAX / a;
	    else limit = LONG_MAX;
	    for (i = 0L;i < n;i++) {
		if (shi >= limit) { puts("エラー: 分子がオーバーフローしました"); return NULL; }
		shi *= a;
	    }
	    bo = 1L;
	    if (b != 0L) limit = LONG_MAX / b;
	    else limit = LONG_MAX;
	    for (i = 0L;i < n;i++) {
		if (bo >= limit) { puts("エラー: 分母がオーバーフローしました"); return NULL; }
		bo *= b;
	    }
	    if (bo == 0L) { puts("エラー: 分母が０です"); return NULL; }
	    return BunsuToCell(shi,bo);
	case 'n':
	    return NumberToCell(pow((long double)a / (long double)b,CellToNumber(y)));
	case 'b':
	    d = pow((long double)a,1.0 / CellToBunbo(y));
	    dd = (long)floor(d);
	    du = (long)ceil(d);
	    n = CellToBunbo(y);
	    k = 1L;
	    for (i = 0;i < n;i++) k *= dd;
	    k2 = 1L;
	    for (i = 0;i < n;i++) k2 *= du;
	    if (k == a || k2 == a) {
		if (k == a) shi = dd; else shi = du;

		d = pow((long double)b,1.0 / CellToBunbo(y));
		dd = (long)floor(d);
		du = (long)ceil(d);
		n = CellToBunbo(y);
		k = 1L;
		for (i = 0;i < n;i++) k *= dd;
		k2 = 1L;
		for (i = 0;i < n;i++) k2 *= du;
		if (k == b || k2 == b) {
		    if (k == b) bo = dd; else bo = du;
		    k = 1L;
		    n = CellToBunshi(y);
		    for (i = 0;i < n;i++) k *= shi;
		    k2 = 1L;
		    for (i = 0;i < n;i++) k2 *= bo;
		    return BunsuToCell(k,k2);
		}
		else {
		    return NumberToCell(pow((long double)a / (long double)b,
			    (long double)CellToBunshi(y) / (long double)CellToBunbo(y) ) );
	        }
	    }
	    else {
		return NumberToCell(pow((long double)a / (long double)b,
			(long double)CellToBunshi(y) / (long double)CellToBunbo(y) ) );
	    }
	default:
	    puts("エラー: 指数部は数値でなければなりません");
	    return NULL;
    }
}

BOOL cmemory::isPlus(CELL *p)
{
    if (p->attr == 'n') {
	if (CellToNumber(p) >= 0.0) return TRUE;
	else return FALSE;
    }
    else if (p->attr == 'i') {
	if (CellToInt(p) >= 0) return TRUE;
	else return FALSE;
    }
    else if (p->attr == 'b') {
	if (CellToBunshi(p) >= 0) return TRUE;
	else return FALSE;
    }
    else return FALSE;
}

BOOL cmemory::isPlusExceptZero(CELL *p)
{
    if (p->attr == 'n') {
	if (CellToNumber(p) > 0.0) return TRUE;
	else return FALSE;
    }
    else if (p->attr == 'i') {
	if (CellToInt(p) > 0) return TRUE;
	else return FALSE;
    }
    else if (p->attr == 'b') {
	if (CellToBunshi(p) > 0) return TRUE;
	else return FALSE;
    }
    else return FALSE;
}

CELL *cmemory::jAbs(CELL *p)
{
    switch (p->attr) {
	case 'n':
	    return NumberToCell(-1.0 * CellToNumber(p));
	case 'i':
	    return IntToCell(-1L * CellToInt(p));
	case 'b':
	    return BunsuToCell(-1L * CellToBunshi(p),CellToBunbo(p));
	default:
	    puts("エラー: 指数部は数値でなければなりません");
	    return NULL;
    }
}

CELL *cfunction::jExpt(CELL *x,CELL *y,SYMVAL *loc,long locn,long lock)
{

    IncCellStack(NULL);
    IncCellStack(y);

    x = eval(x,loc,locn,lock);

    PopCellStack();

    if (x == NULL) return NULL;

    IncCellStack(NULL);
    IncCellStack(x);

    y = eval(y,loc,locn,lock);

    PopCellStack();

    if (y == NULL) return NULL;

    if (!isPlus(x)) {
	puts("エラー: 正数が必要です");
	errorlisp(loc,locn);
	return NULL;
    }

    if (x->attr == 'i' && isPlus(y)) {
	return ijExpt(CellToInt(x),y);
    }
    else if (x->attr == 'n') {
	return njExpt(CellToNumber(x),y);
    }
    else if (x->attr == 'b' && isPlus(y)) {
	return bjExpt(CellToBunshi(x),CellToBunbo(x),y);
    }
    else if (x->attr == 'i') {
	y = jAbs(y);
	if (y == NULL) return NULL;
	return bjExpt(1L,CellToInt(x),y);
    }
    else if (x->attr == 'b' && CellToBunshi(x) == 1L) {
	y = jAbs(y);
	if (y == NULL) return NULL;
	return ijExpt(CellToBunbo(x),y);
    }
    else if (x->attr == 'b') {
	y = jAbs(y);
	if (y == NULL) return NULL;
	return bjExpt(CellToBunbo(x),CellToBunshi(x),y);
    }
    else {
	puts("エラー: 数値のべき乗ではありません");
	errorlisp(loc,locn);
	return NULL;
    }
}

CELL *cfunction::jExp(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 0L) return IntToCell(1L);
	else {
	    return NumberToCell(exp((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(exp(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(exp((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("エラー: EXP 数値の引数が必要です");
	errorlisp(loc,locn);
	return NULL;
    }
}

CELL *cfunction::jLog(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (!isPlusExceptZero(p)) {
	puts("エラー: 正数が必要です");
	errorlisp(loc,locn);
	return NULL;
    }
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 1L) return IntToCell(0L);
	else {
	    return NumberToCell(log((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(log(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(log((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("エラー: LOG 数値の引数が必要です");
	errorlisp(loc,locn);
	return NULL;
    }
}

CELL *cfunction::jSin(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 0L) return IntToCell(0L);
	else {
	    return NumberToCell(sin((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(sin(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(sin((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("エラー: SIN 数値の引数が必要です");
	errorlisp(loc,locn);
	return NULL;
    }
}

CELL *cfunction::jCos(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 0L) return IntToCell(1L);
	else {
	    return NumberToCell(cos((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(cos(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(cos((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("エラー: COS 数値の引数が必要です");
	errorlisp(loc,locn);
	return NULL;
    }
}

CELL *cfunction::jTan(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 0L) return IntToCell(0L);
	else {
	    return NumberToCell(tan((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(tan(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(tan((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("エラー: TAN 数値の引数が必要です");
	errorlisp(loc,locn);
	return NULL;
    }
}

BOOL cmemory::isInRange(CELL *p)
{
    long i;
    long double v;

    switch (p->attr) {
	case 'i':
	    i = CellToInt(p);
	    if (-1L <= i && i <= 1L) return TRUE;
	    else return FALSE;
	case 'n':
	    v = CellToNumber(p);
	    if (-1.0 <= v && v <= 1.0) return TRUE;
	    else return FALSE;
	case 'b':
	    v = (long double)CellToBunshi(p) / (long double)CellToBunbo(p);
	    if (-1.0 <= v && v <= 1.0) return TRUE;
	    else return FALSE;
	default:
	    return FALSE;
    }
}

CELL *cfunction::jAsin(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long i;
    long double v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (!isInRange(p)) {
	puts("エラー: 範囲外です");
	errorlisp(loc,locn);
	return NULL;
    }
    switch (p->attr) {
	case 'i':
	    i = CellToInt(p);
	    if (i == 0L) return IntToCell(0L);
	    return NumberToCell(asin((long double)i));
	case 'n':
	    return NumberToCell(asin(CellToNumber(p)));
	case 'b':
	    v = (long double)CellToBunshi(p) / (long double)CellToBunbo(p);
	    return NumberToCell(asin(v));
	default:
	    puts("エラー: 数値ではありません");
	    errorlisp(loc,locn);
	    return NULL;
    }
}

CELL *cfunction::jAcos(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long i;
    long double v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (!isInRange(p)) {
	puts("エラー: 範囲外です");
	errorlisp(loc,locn);
	return NULL;
    }
    switch (p->attr) {
	case 'i':
	    i = CellToInt(p);
	    if (i == 1L) return IntToCell(0L);
	    return NumberToCell(acos((long double)i));
	case 'n':
	    return NumberToCell(acos(CellToNumber(p)));
	case 'b':
	    v = (long double)CellToBunshi(p) / (long double)CellToBunbo(p);
	    return NumberToCell(acos(v));
	default:
	    puts("エラー: 数値ではありません");
	    errorlisp(loc,locn);
	    return NULL;
    }
}

CELL *cfunction::jAtan(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long i;
    long double v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    switch (p->attr) {
	case 'i':
	    i = CellToInt(p);
	    if (i == 0L) return IntToCell(0L);
	    return NumberToCell(atan((long double)i));
	case 'n':
	    return NumberToCell(atan(CellToNumber(p)));
	case 'b':
	    v = (long double)CellToBunshi(p) / (long double)CellToBunbo(p);
	    return NumberToCell(atan(v));
	default:
	    puts("エラー: 数値ではありません");
	    errorlisp(loc,locn);
	    return NULL;
    }
}

CELL *cfunction::jEvenp(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v,w,x;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr != 'i') {
	puts("エラー: 整数ではありません");
	errorlisp(loc,locn);
	return NULL;
    }
    v = CellToInt(p);
    w = v / 2L;
    x = v - w * 2L;
    if (x == 1L || x == -1L) return NILCell;
    return TCell;
}

CELL *cfunction::tfloor(CELL *p,SYMVAL *loc,long locn,long lock)
{
	long double f;
	AINT *k;
	long k1,k2;
	CELL *q,*a,*r,*r2,*q2;
	AFLOAT *n;

	p = eval(p,loc,locn,lock);
	if (p == NULL) return NULL;
	if (p->attr == 'i') {
	    q = getCell();
	    *q = *p;
	    k = getInt(0L);
	    a = getCell();
	    a->attr = 'i';
	    a->car = (CELL *)k;
	    r = getCell();
	    r2 = getCell();
	    r->attr = 'l';
	    r->car = q;
	    r->cdr = r2;
	    r2->attr = 'l';
	    r2->car = a;
	    r2->cdr = NILCell;
	    q->mtop = r;
	    return q;
	}
	else if (p->attr == 'n') {
	    q = getCell();
	    k1 = (long)floor((double)(((AFLOAT *)(p->car))->f));
	    k = getInt(k1);
	    q->attr = 'i';
	    q->car = (CELL *)k;
	    f = ((AFLOAT *)(p->car))->f - k1;
	    n = getNumber(f);
	    a = getCell();
	    a->attr = 'n';
	    a->car = (CELL *)n;
	    r = getCell();
	    r2 = getCell();
	    r->attr = 'l';
	    r->car = q;
	    r->cdr = r2;
	    r2->attr = 'l';
	    r2->car = a;
	    r2->cdr = NILCell;
	    q->mtop = r;
	    return q;
	}
	else if (p->attr == 'b') {
	    q = getCell();
	    k1 = CellToBunshi(p);
	    k2 = CellToBunbo(p);
	    f = (long double)k1 / (long double)k2;
	    k1 = (long)floor((double)f);
	    k = getInt(k1);
	    q->attr = 'i';
	    q->car = (CELL *)k;
	    q2 = getCell();
	    *q2 = *q;
	    if ((a = Sa(p,q2)) == NULL) {
		errorlisp(loc,locn);
		return NULL;
	    }
	    r = getCell();
	    r2 = getCell();
	    r->attr = 'l';
	    r->car = q;
	    r->cdr = r2;
	    r2->attr = 'l';
	    r2->car = a;
	    r2->cdr = NILCell;
	    q->mtop = r;
	    return q;
	}
	else {
	    puts("エラー: 数値ではありません");
	    errorlisp(loc,locn);
	    return NULL;
	}
}

CELL *cmemory::now(void)
{
    time_t t;
    CELL *w,*ca,*cd;
    AINT *ia,*id;

    t = time(NULL);
    w = getCell();
    ca = getCell();
    cd = getCell();
    ia = getInt((long)(t / 3600L));
    id = getInt((long)(t % 3600L));
    w->attr = 'l';
    ca->attr = 'i';
    cd->attr = 'i';
    ca->car = (CELL *)ia;
    cd->car = (CELL *)id;
    w->car = ca;
    w->cdr = cd;
    return w;
}

CELL *cfunction::LocalTime(CELL *p,SYMVAL *loc,long locn,long lock)
{
    CELL *x1,*x2,*x3,*x4,*x5,*x6;
    CELL *sec,*min,*hour,*mday,*mon,*year,*wday;
    AINT *sec_i,*min_i,*hour_i,*mday_i,*mon_i,*year_i,*wday_i;
    long a,b,c;
    struct tm *d,e;

    p = eval(p,loc,locn,lock);
    if (p == NULL) goto aerror;
    if ((p->car)->attr != 'i') goto aerror;
    if ((p->cdr)->attr != 'i') goto aerror;
    a = ((AINT *)((p->car)->car))->f;
    b = ((AINT *)((p->cdr)->car))->f;
    c = a * 3600L + b;
    d = localtime(&c);
    if (d == NULL) goto aerror;
    e.tm_sec = d->tm_sec;
    e.tm_min = d->tm_min;
    e.tm_hour = d->tm_hour;
    e.tm_mday = d->tm_mday;
    e.tm_mon = d->tm_mon + 1;
    e.tm_year = d->tm_year + 1900;
    e.tm_wday = d->tm_wday;
    sec_i = getInt(e.tm_sec);
    sec = getCell();
    sec->attr = 'i';
    sec->car = (CELL *)sec_i;
    min_i = getInt(e.tm_min);
    min = getCell();
    min->attr = 'i';
    min->car = (CELL *)min_i;
    hour_i = getInt(e.tm_hour);
    hour = getCell();
    hour->attr = 'i';
    hour->car = (CELL *)hour_i;
    mday_i = getInt(e.tm_mday);
    mday = getCell();
    mday->attr = 'i';
    mday->car = (CELL *)mday_i;
    mon_i = getInt(e.tm_mon);
    mon = getCell();
    mon->attr = 'i';
    mon->car = (CELL *)mon_i;
    year_i = getInt(e.tm_year);
    year = getCell();
    year->attr = 'i';
    year->car = (CELL *)year_i;
    wday_i = getInt(e.tm_wday);
    wday = getCell();
    wday->attr = 'i';
    wday->car = (CELL *)wday_i;
    x1 = getCell();
    x1->attr = 'l';
    x1->car = year;
    x1->cdr = wday;
    x2 = getCell();
    x2->attr = 'l';
    x2->car = mon;
    x2->cdr = x1;
    x3 = getCell();
    x3->attr = 'l';
    x3->car = mday;
    x3->cdr = x2;
    x4 = getCell();
    x4->attr = 'l';
    x4->car = hour;
    x4->cdr = x3;
    x5 = getCell();
    x5->attr = 'l';
    x5->car = min;
    x5->cdr = x4;
    x6 = getCell();
    x6->attr = 'l';
    x6->car = sec;
    x6->cdr = x5;
    return x6;

aerror:
    puts("エラー: LOCAL-TIMEの引数にエラーがあります。");
    errorlisp(loc,locn);
    return NULL;
}

CELL *cfunction::MakeTime(CELL *p,SYMVAL *loc,long locn,long lock)
{
    struct tm d;
    time_t t;
    CELL *w,*ca,*cd;
    AINT *ia,*id;

    p = eval(p,loc,locn,lock);
    if (p == NULL) goto aerror;
    if (p->attr != 'l') goto aerror;
    if ((p->car)->attr != 'i') goto aerror;
    d.tm_sec = ((AINT *)((p->car)->car))->f;
    p = p->cdr;
    if (p->attr != 'l') goto aerror;
    if ((p->car)->attr != 'i') goto aerror;
    d.tm_min = ((AINT *)((p->car)->car))->f;
    p = p->cdr;
    if (p->attr != 'l') goto aerror;
    if ((p->car)->attr != 'i') goto aerror;
    d.tm_hour = ((AINT *)((p->car)->car))->f;
    p = p->cdr;
    if (p->attr != 'l') goto aerror;
    if ((p->car)->attr != 'i') goto aerror;
    d.tm_mday = ((AINT *)((p->car)->car))->f;
    p = p->cdr;
    if (p->attr != 'l') goto aerror;
    if ((p->car)->attr != 'i') goto aerror;
    d.tm_mon = ((AINT *)((p->car)->car))->f - 1;
    p = p->cdr;
    if (p->attr != 'l') goto aerror;
    if ((p->car)->attr != 'i') goto aerror;
    d.tm_year = ((AINT *)((p->car)->car))->f - 1900;
    d.tm_isdst = -1;

    t = mktime(&d);
    if (t == -1) goto aerror;
    w = getCell();
    ca = getCell();
    cd = getCell();
    ia = getInt((long)(t / 3600L));
    id = getInt((long)(t % 3600L));
    w->attr = 'l';
    ca->attr = 'i';
    cd->attr = 'i';
    ca->car = (CELL *)ia;
    cd->car = (CELL *)id;
    w->car = ca;
    w->cdr = cd;
    return w;

aerror:
    puts("エラー: MAKE-TIMEの引数にエラーがあります。");
    errorlisp(loc,locn);
    return NULL;
}

// end of file

